##
# File: FileLock.py
# Date: 3-Sep-2021
#
# A simple implementation of locking using a file system lock file.
# Adapted from a portion of the py-filelock package https://github.com/benediktschmitt/py-filelock.
##

__docformat__ = "google en"
__author__ = "John Westbrook"
__email__ = "jwest@rcsb.rutgers.edu"
__license__ = "Apache 2.0"

import logging
import os
import threading
import time
import typing

from rcsb.utils.io.FileUtil import FileUtil

logger = logging.getLogger(__name__)


class Timeout(TimeoutError):
    pass


class AcquireProxy(object):
    """This proxy class avoids invoking the __enter__() method multiple times."""

    def __init__(self, lock):
        self.lock = lock
        return None

    def __enter__(self):
        return self.lock

    def __exit__(self, excType, excValue, traceback):
        self.lock.release()
        return None


class FileLock(object):
    """Implements a locking protocol based on the existence of a file."""

    def __init__(self, lockFilePath: str, defaultTimeout: typing.Optional[float] = 5.0):
        """Implements a locking protocol based on the existence of a file.

        Args:
            lockFilePath (str): path to lock file
            defaultTimeout (typing.Optional[float], optional): time out to acquire a lock (seconds). Defaults to 5.0.

        """

        # The path to the lock file.
        self.__lockFilePath = lockFilePath
        #
        # File descriptor for the lock file returned by os.open() and None indicates that no lock is held.
        self.__lockFileFileDescriptor = None

        # The default timeout value.
        self.__defaultTimeout = defaultTimeout

        # Protects the lock count
        self.__lockCountLock = threading.Lock()

        # Lock count to manage nested context manager instances of locking
        self.__lockCount = 0
        return None

    def getLockFilePath(self) -> str:
        """Return the path to the lock file."""
        return self.__lockFilePath

    def isLocked(self) -> bool:
        """Return the status of the file lock."""
        return self.__lockFileFileDescriptor is not None

    def acquire(self, timeout: typing.Optional[float] = None, pollInterval: typing.Optional[float] = 0.05) -> typing.Type[AcquireProxy]:
        """Acquire file lock.

        Args:
            timeout (typing.Optional[float], optional): timeout to acquire lock (seconds). Defaults to None.
            pollInterval (typing.Optional[float], optional): retry interval (seconds). Defaults to 0.05.

        Raises:
            Timeout: if acquire timeout is exceeded

        Returns:
            AcquireProxy: returns a proxy object for the file lock.
        """

        # Use the default timeout, if no timeout is provided.
        timeout = timeout if timeout else self.__defaultTimeout

        with self.__lockCountLock:
            self.__lockCount += 1

        lockId = id(self)
        lockFilePath = self.__lockFilePath
        startTime = time.time()
        try:
            while True:
                with self.__lockCountLock:
                    if not self.isLocked():
                        logger.debug("Attempting to acquire lock %s on %s", lockId, lockFilePath)
                        self.__doAquireLock()

                if self.isLocked():
                    logger.debug("Lock %s acquired on %s", lockId, lockFilePath)
                    break
                elif timeout >= 0 and time.time() - startTime > timeout:
                    logger.debug("Timeout on acquiring lock %s on %s", lockId, lockFilePath)
                    raise Timeout(self.__lockFilePath)
                else:
                    logger.debug("Lock %s not acquired on %s, waiting %s seconds ...", lockId, lockFilePath, pollInterval)
                    time.sleep(pollInterval)
        except Exception:
            with self.__lockCountLock:
                self.__lockCount = max(0, self.__lockCount - 1)

            raise
        return AcquireProxy(lock=self)

    def release(self, force: typing.Optional[bool] = False):
        """Release an acquired file lock.

        Args:
            force (typing.Optional[bool], optional): Force release even if there are lock holders. Defaults to False.
        """
        with self.__lockCountLock:
            if self.isLocked():
                self.__lockCount -= 1

                if self.__lockCount == 0 or force:
                    lockId = id(self)
                    logger.debug("Attempting to release lock %s on %s", lockId, self.__lockFilePath)
                    self.__doReleaseLock()
                    self.__lockCount = 0
                    logger.debug("Lock %s released on %s", lockId, self.__lockFilePath)
        return None

    def __doAquireLock(self):
        fU = FileUtil()
        mode = os.O_WRONLY | os.O_CREAT | os.O_EXCL | os.O_TRUNC
        try:
            fU.mkdir(os.path.dirname(self.__lockFilePath))
            fd = os.open(self.__lockFilePath, mode)
        except (IOError, OSError):
            pass
        else:
            self.__lockFileFileDescriptor = fd
        return None

    def __doReleaseLock(self):
        os.close(self.__lockFileFileDescriptor)
        self.__lockFileFileDescriptor = None
        try:
            os.remove(self.__lockFilePath)
        except OSError:
            pass
        return None

    def __enter__(self):
        self.acquire()
        return self

    def __exit__(self, excType, excValue, traceback):
        self.release()
        return None

    def __del__(self):
        self.release(force=True)
        return None



##
# File: FastaUtil.py
# Date:  24-Nov-2018 Jdw
#

import logging
import re
from gzip import GzipFile

logger = logging.getLogger(__name__)


class FastaUtil(object):
    """Simple FASTA reader and writer with options to parse application
    specific common lines (e.g. Uniprot and PDB)
    """

    aaDict3 = {
        "ALA": "A",
        "ARG": "R",
        "ASN": "N",
        "ASP": "D",
        "ASX": "B",
        "CYS": "C",
        "GLN": "Q",
        "GLU": "E",
        "GLX": "Z",
        "GLY": "G",
        "HIS": "H",
        "ILE": "I",
        "LEU": "L",
        "LYS": "K",
        "MET": "M",
        "PHE": "F",
        "PRO": "P",
        "SER": "S",
        "THR": "T",
        "TRP": "W",
        "TYR": "Y",
        "VAL": "V",
        "PYL": "O",
        "SEC": "U",
    }
    # aaValidCodes = "".join(sorted(FastaUtil.aaDict3.values()))
    aaValidCodes = "ABCDEFGHIKLMNOPQRSTUVWYZX"
    naValidCodes = "AGCTURYNWSMKBHDV"

    def __init__(self, **kwargs):
        self.__uniProtCommentRegex = re.compile(r"\>(\w+)\|(\w+)\|(\w+)\s(.*)\sOS=(.*)\sOX=(\d+)\s(GN=(.*)\s)?PE=(\d+)\sSV=(\d+)")
        _ = kwargs

    def __removeWhiteSpace(self, string):
        pattern = re.compile(r"\s+")
        return re.sub(pattern, "", string)

    def cleanSequence(self, sequence, seqType="protein"):
        """[summary]

        Args:
            sequence ([type]): raw one-letter code sequence
            seqType (str, optional): sequence type protein or na ... Defaults to "protein".

        Returns:
            (bool, str): status, cleaned sequence (upper case w/o white space)
        """
        try:
            tS = self.__removeWhiteSpace(sequence.upper())
            if seqType == "protein":
                allowedChars = set(FastaUtil.aaValidCodes)
            else:
                allowedChars = set(FastaUtil.naValidCodes)

            if set(sequence).issubset(allowedChars):
                return True, tS
        except Exception as e:
            logger.exception("Failing with %s", str(e))
        return False, sequence

    def parseComment(self, cmtLine, commentStyle):
        if commentStyle == "uniprot":
            seqId, cD = self.__parseCommentUniProt(cmtLine)
        elif commentStyle == "uniprot_regex":
            seqId, cD = self.__parseCommentUniProtRe(cmtLine)
        elif commentStyle == "prerelease":
            seqId, cD = self.__parseCommentPreRelease(cmtLine)
        elif commentStyle == "tagged":
            seqId, cD = self.__parseCommentTagged(cmtLine)
        else:
            seqId, cD = self.__parseCommentDefault(cmtLine)
        return seqId, cD

    def readFasta(self, filePath, **kwargs):
        """"""
        try:
            commentStyle = kwargs.get("commentStyle", "uniprot").lower()
            if commentStyle == "uniprot":
                commentParser = self.__parseCommentUniProt
            elif commentStyle == "uniprot_regex":
                commentParser = self.__parseCommentUniProtRe
            elif commentStyle == "prerelease":
                commentParser = self.__parseCommentPreRelease
            elif commentStyle == "tagged":
                commentParser = self.__parseCommentTagged
            else:
                commentParser = self.__parseCommentDefault

            if filePath[-3:] == ".gz":
                sD = self.__readFastaGz(filePath, commentParser)
            else:
                sD = self.__readFasta(filePath, commentParser)

            return sD

        except Exception as e:
            logger.exception("Failing for filePath %r with %s", filePath, str(e))

        return {}

    def writeFasta(self, filePath, seqDict, **kwargs):
        """"""
        maxLineLength = int(kwargs.get("maxLineLength", 70))
        makeComment = kwargs.get("makeComment", False)
        with open(filePath, "w") as ofh:
            ok = self.__writeFasta(ofh, seqDict, maxLineLength=maxLineLength, makeComment=makeComment)
        return ok

    def __readFastaGz(self, filePath, commentParser):
        seqD = {}
        with GzipFile(filePath, "rb") as ifh:
            for cmtLine, sequence in self.__readRecordFasta(ifh):
                seqId, cD = commentParser(cmtLine)
                cD["sequence"] = sequence.upper()
                if seqId:
                    seqD[seqId] = cD
        return seqD

    def __readFasta(self, filePath, commentParser):
        seqD = {}
        with open(filePath, "r") as ifh:
            for cmtLine, sequence in self.__readRecordFasta(ifh):
                seqId, cD = commentParser(cmtLine)
                cD["sequence"] = sequence.upper()
                if seqId:
                    seqD[seqId] = cD
        return seqD

    def __parseCommentUniProt(self, cmtLine):
        """
        >sp|O00141|SGK1_HUMAN Serine/threonine-protein kinase Sgk1 OS=Homo sapiens OX=9606 GN=SGK1 PE=1 SV=2
        """
        try:
            org = ""
            geneName = ""
            dbName = ""
            dbAccession = ""
            dbIsoform = ""
            seqId = ""
            description = ""
            #
            ff = cmtLine[1:].split("|")
            dbName = ff[0].upper()
            seqId = ff[1]
            tS = ff[2]
            #
            tt = seqId.split("-")
            dbAccession = seqId
            dbIsoform = ""
            if len(tt) > 1:
                dbAccession = tt[0]
                dbIsoform = tt[1]
            #
            idx = tS.find("OS=")
            if idx > 0:
                description = tS[: idx - 1]
                ff = tS[idx:].split("=")
                if len(ff) == 3:
                    if ff[0] == "OS":
                        org = ff[1][:-2]
                        geneName = ff[2]
                elif len(ff) > 1:
                    if ff[0] == "OS":
                        org = ff[1]
            #
            cD = {"description": description, "org": org, "gene_name": geneName, "db_accession": dbAccession, "db_name": dbName, "db_isoform": dbIsoform}
            return seqId, cD
            #
        except Exception as e:
            logger.exception("Failed to parse comment %s with %s", cmtLine, str(e))
        return None, {}

    def __parseCommentUniProtRe(self, cmtLine):
        """
        '>sp|O00141|SGK1_HUMAN Serine/threonine-protein kinase Sgk1 OS=Homo sapiens OX=9606 GN=SGK1 PE=1 SV=2
        """
        try:
            org = ""
            geneName = ""
            dbName = ""
            dbAccession = ""
            dbIsoform = ""
            seqId = ""
            description = ""
            taxId = ""
            # proteinExistence = ""
            # seqVersion = ""
            #
            match = self.__uniProtCommentRegex.match(cmtLine)
            if match:
                groups = match.groups()
                dbName = groups[0]
                seqId = groups[1]
                tt = seqId.split("-")
                dbAccession = seqId
                dbIsoform = ""
                if len(tt) > 1:
                    dbAccession = tt[0]
                    dbIsoform = tt[1]
                entryName = groups[2]
                description = groups[3]
                org = groups[4]
                taxId = groups[5]
                geneName = groups[7]
                # proteinExistence = groups[8]
                # seqVersion = groups[9]
            #
            cD = {
                "seqId": seqId,
                "description": description,
                "org": org,
                "taxonomy_id": taxId,
                "gene_name": geneName,
                "db_accession": dbAccession,
                "entry_name": entryName,
                "db_name": dbName,
                "db_isoform": dbIsoform,
                # "protein_existence": proteinExistence,
                # "seq_version": seqVersion,
            }
            return seqId, cD
            #
        except Exception as e:
            logger.exception("Failed to parse comment %s with %s", cmtLine, str(e))
        return None, {}

    def __parseCommentTagged(self, cmtLine):
        try:
            cD = {}
            fL = cmtLine[1:].split("|")
            for v, k in zip(fL[0::2], fL[1::2]):
                cD[k] = v
            seqId = cD["seqId"]
            return seqId, cD
            #
        except Exception as e:
            logger.exception("Failed to parse comment %s with %s", cmtLine, str(e))
        return None, {}

    def __parseCommentPreRelease(self, cmtLine):
        try:
            #
            ff = cmtLine[1:].split(" ")
            entryId = ff[0].upper()
            entityId = ff[2]
            seqId = entryId + "_" + entityId
            return seqId, {}
            #
        except Exception as e:
            logger.exception("Failed to parse comment %s with %s", cmtLine, str(e))
        return None, {}

    def __parseCommentDefault(self, cmtLine):
        try:
            #
            seqId = cmtLine[1:]
            return seqId, {}
            #
        except Exception as e:
            logger.exception("Failed to parse comment %s with %s", cmtLine, str(e))
        return None, {}

    def __readRecordFasta(self, ifh):
        """Return the next FASTA record in the input file handle."""
        comment, sequence = None, []
        for line in ifh:
            # handle the variety of types accross versions and open/gzipfile classes
            if isinstance(line, str):
                line = line[:-1].encode("ascii", "xmlcharrefreplace").decode("ascii")
            else:
                line = line[:-1].decode("ascii", "xmlcharrefreplace")

            line = line.rstrip()
            if line.startswith(">"):
                if comment:
                    yield (comment, "".join(sequence))
                comment, sequence = line, []
            else:
                sequence.append(line)
        if comment:
            yield (comment, "".join(sequence))

    def __writeFasta(self, ofh, seqDict, maxLineLength=70, makeComment=False):
        """
        seqDict[seqId] = {sequence: 'one-letter-code sequence', ...}

        With makeComment = False:

            >seqId
            MTVEDR....

        or makeComment= True :

            >{v|k|v'|k' ... for k,v in seqD.items()}
            MTVEDR....

        """
        try:
            lCount = 0
            for seqId, sD in seqDict.items():
                sA = []
                if makeComment:
                    comment = ""
                    cL = []
                    if "seqId" in sD:
                        cL.append(str(sD["seqId"]))
                        cL.append("seqId")
                    for k, v in sD.items():
                        if k in ["sequence", "seqId"]:
                            continue
                        cL.append(str(v))
                        cL.append(str(k))
                    comment = "|".join(cL)
                else:
                    comment = seqId
                #
                sA.append(">%s\n" % comment)
                lCount += 1
                sequence = sD["sequence"]
                for i in range(0, len(sequence), maxLineLength):
                    sA.append(sequence[i: i + maxLineLength] + "\n")
                    result = "".join(sA)
                    lCount += 1
                ofh.write(result)
            return True
        except Exception as e:
            logger.info("seqId %r sD %r", seqId, sD)
            logger.exception("Failing with %s", str(e))

        return False

    def __readFastaX(self, filePath):
        try:
            rD = {}
            with open(filePath, "r") as fb:
                for line in fb:
                    if str(line[:-1]).startswith(">"):
                        fields = str(line[:-1]).strip().split()
                        ky = str(fields[0][1:]).strip()
                        length = int(str(fields[2][7:]))
                        rD[ky] = length
                return rD
        except Exception as e:
            logger.exception("Failing wiht %s", str(e))
        return rD



##
# File: StashUtil.py
#
# Utilities to stash and recover a data in collection of directories to and from
# remote sftp, http or local POSIX file storage resources.
#
# Updates:
# 19-Jul-2021 jdw add git push support
#
##

__docformat__ = "google en"
__author__ = "John Westbrook"
__email__ = "jwest@rcsb.rutgers.edu"
__license__ = "Apache 2.0"

import logging
import os

from rcsb.utils.io.FileUtil import FileUtil
from rcsb.utils.io.GitUtil import GitUtil
from rcsb.utils.io.SftpUtil import SftpUtil
from rcsb.utils.io.SplitJoin import SplitJoin

logger = logging.getLogger(__name__)


class StashUtil(object):
    """Utilities to stash and recover a data in collection of (sub)directories to/from
    remote sftp, http or local POSIX file storage resources.
    """

    def __init__(self, localBundlePath, baseBundleFileName):
        """Set the bundle file name and path for class instance.

        Args:
            localBundlePath (str): writeable local path for bundle file artifact
            baseBundleFileName (str): bundle file name without file extension
        """
        #
        self.__localBundlePath = localBundlePath
        self.__baseBundleName = baseBundleFileName
        self.__baseBundleFileName = self.__baseBundleName + ".tar.gz"
        self.__localStashTarFilePath = os.path.join(localBundlePath, self.__baseBundleFileName)

    def makeBundle(self, localParentPath, subDirList):
        """Bundle the subdirectories of the input parent directory path.

        Args:
            localParentPath (str): local parent directory path containing the bundling targets
            subDirList (list, str): list of subdirectories of the parent path to be bundled

        Returns:
            (bool): True for success or False otherwise
        """
        fileU = FileUtil()
        fileU.mkdir(self.__localBundlePath)
        dirPathList = [os.path.join(localParentPath, subDir) for subDir in subDirList]
        okT = fileU.bundleTarfile(self.__localStashTarFilePath, dirPathList, mode="w:gz", recursive=True)
        return okT

    def storeBundle(self, url, remoteDirPath, remoteStashPrefix="A", userName=None, password=None):
        """Store a copy of the bundled search dependencies remotely -

        Args:
            url (str): URL string for the destination host (e.g. sftp://myserver.net or None for a local file)
            remoteDirPath (str): remote directory path on the remote resource
            remoteStashPrefix (str, optional): optional label preppended to the stashed dependency bundle artifact (default='A')
            userName (str, optional): optional access information. Defaults to None.
            password (str, optional): optional access information. Defaults to None.

        Returns:
          bool:  True for success or False otherwise

        """
        try:
            ok = False
            fn = self.__makeBundleFileName(self.__baseBundleFileName, remoteStashPrefix=remoteStashPrefix)
            if url and url.startswith("sftp://"):
                sftpU = SftpUtil()
                hostName = url[7:]
                ok = sftpU.connect(hostName, userName, pw=password, port=22)
                if ok:
                    remotePath = os.path.join("/", remoteDirPath, fn)
                    ok = sftpU.put(self.__localStashTarFilePath, remotePath)
            elif not url:
                fileU = FileUtil()
                remotePath = os.path.join(remoteDirPath, fn)
                ok = fileU.put(self.__localStashTarFilePath, remotePath)
            else:
                logger.error("Unsupported stash protocol %r", url)
            return ok
        except Exception as e:
            logger.exception("For %r %r failing with %s", url, remoteDirPath, str(e))
        return False

    def fetchBundle(self, localRestoreDirPath, url, remoteDirPath, remoteStashPrefix="A", userName=None, password=None):
        """Restore bundled dependencies from remote storage and unbundle these in the
           current local cache directory.

        Args:
            localRestoreDirPath (str): local restore path
            url (str): remote URL
            remoteDirPath (str): remote directory path on the remote resource
            remoteStashPrefix (str, optional): optional label preppended to the stashed dependency bundle artifact (default='A')
            userName (str, optional): optional access information. Defaults to None.
            password (str, optional): optional access information. Defaults to None.
        """
        try:
            ok = False
            fileU = FileUtil()
            fn = self.__makeBundleFileName(self.__baseBundleFileName, remoteStashPrefix=remoteStashPrefix)
            if not url:
                remotePath = os.path.join(remoteDirPath, fn)
                if fileU.exists(remotePath):
                    ok = fileU.get(remotePath, self.__localStashTarFilePath)
                else:
                    ok = False
                    logger.warning("Missing bundle file %r", remotePath)

            elif url and (url.startswith("http://") or url.startswith("https://")):
                remotePath = url + os.path.join("/", remoteDirPath, fn)
                ok = fileU.get(remotePath, self.__localStashTarFilePath)

            elif url and url.startswith("sftp://"):
                sftpU = SftpUtil()
                ok = sftpU.connect(url[7:], userName, pw=password, port=22)
                if ok:
                    remotePath = os.path.join(remoteDirPath, fn)
                    ok = sftpU.get(remotePath, self.__localStashTarFilePath)
            else:
                logger.error("Unsupported protocol %r", url)
            if ok:
                ok = fileU.unbundleTarfile(self.__localStashTarFilePath, dirPath=localRestoreDirPath)
            return ok
        except Exception as e:
            logger.exception("For %r %r Failing with %s", url, remoteDirPath, str(e))
            ok = False
        return ok

    def pushBundle(self, gitRepositoryPath, accessToken, gitHost="github.com", gitBranch="master", remoteStashPrefix="A", maxSizeMB=95):
        """Push bundle to remote stash git repository.

        Args:
            gitRepositoryPath (str): git repository path (e.g., rcsb/py-rcsb_exdb_assets_stash)
            accessToken (str): git repository access token
            gitHost (str, optional): git repository host name. Defaults to github.com.
            gitBranch (str, optional): git branch name. Defaults to master.
            remoteStashPrefix (str, optional): optional label preppended to the stashed dependency bundle artifact (default='A')
            maxSizeMB (int, optional): maximum stash bundle file size that will be committed. Defaults to 95MB.

        Returns:
          bool:  True for success or False otherwise

        """
        try:
            ok = False
            gU = GitUtil(token=accessToken, repositoryHost=gitHost)
            fU = FileUtil()
            localRepositoryPath = os.path.join(self.__localBundlePath, "stash_repository")
            fn = self.__makeBundleFileName(self.__baseBundleFileName, remoteStashPrefix=remoteStashPrefix)
            #
            # Update existing local repository, otherwise clone a new copy
            if fU.exists(localRepositoryPath):
                ok = gU.pull(localRepositoryPath, branch=gitBranch)
                logger.debug("After pull status %r", gU.status(localRepositoryPath))
            else:
                ok = gU.clone(gitRepositoryPath, localRepositoryPath, branch=gitBranch)
            #
            # Split all bundles
            mbSize = float(fU.size(self.__localStashTarFilePath)) / 1000000.0
            logger.info("Splitting bundle %r (%.3f MB/Max %d MB)", fn, mbSize, maxSizeMB)
            sj = SplitJoin()
            splitDirPath = os.path.join(localRepositoryPath, "stash", fn[:-7])
            sj.split(self.__localStashTarFilePath, splitDirPath, maxSizeMB=maxSizeMB)
            fU.remove(self.__localStashTarFilePath)
            # else:
            # fU.put(self.__localStashTarFilePath, os.path.join(localRepositoryPath, "stash", fn))

            ok = gU.addAll(localRepositoryPath, branch=gitBranch)
            ok = gU.commit(localRepositoryPath, branch=gitBranch)
            logger.debug("After commit status %r", gU.status(localRepositoryPath))
            #
            if accessToken:
                ok = gU.push(localRepositoryPath, branch=gitBranch)
                logger.info("After push status %r", gU.status(localRepositoryPath))
            #
            return ok
        except Exception as e:
            logger.exception("For %r %r failing with %s", gitHost, gitRepositoryPath, str(e))
        return False

    def fetchPartitionedBundle(self, localRestoreDirPath, gitRepositoryPath, gitRawHost="raw.githubusercontent.com", gitBranch="master", remoteStashPrefix="A"):
        """Fetch bundle from a remote stash public git repository via http.

        Args:
            localRestoreDirPath (str): local restore path
            gitRepositoryPath (str): git repository path (e.g., rcsb/py-rcsb_exdb_assets_stash)
            gitHost (str, optional): git repository host name. Defaults to github.com.
            gitBranch (str, optional): git branch name. Defaults to master.
            remoteStashPrefix (str, optional): optional label preppended to the stashed dependency bundle artifact (default='A')

        Returns:
          bool:  True for success or False otherwise

            https://raw.githubusercontent.com/rcsb/py-rcsb_exdb_assets_stash/master/stash/<file_or_dir>
        """
        try:
            ok = False
            fileU = FileUtil()
            bundleFileName = self.__makeBundleFileName(self.__baseBundleFileName, remoteStashPrefix=remoteStashPrefix)
            urlBase = "https://" + gitRawHost
            rp = gitRepositoryPath[:-4] if gitRepositoryPath.endswith(".git") else gitRepositoryPath
            repoDirPath = os.path.join(urlBase, rp, gitBranch, "stash")

            # First fetch the manifest file
            remoteDirPath = os.path.join(repoDirPath, bundleFileName[:-7])
            remotePath = os.path.join(remoteDirPath, "MANIFEST")
            logger.debug("Manifest remote %r", remotePath)
            #
            localDirPath = os.path.join(self.__localBundlePath, bundleFileName[:-7])
            manifestPath = os.path.join(localDirPath, "MANIFEST")

            ok = fileU.get(remotePath, manifestPath)
            if not ok:
                logger.error("No manifest file at %r", remotePath)
                return ok
            # ---
            partFileName = "part_1"
            remotePartPath = os.path.join(repoDirPath, bundleFileName[:-7], partFileName)
            logger.debug("remotePartPath %r", remotePartPath)
            # ---
            partList = []
            with open(manifestPath, "r") as mfh:
                line = mfh.readline()
                tf, myHash = line[:-1].split("\t")
                logger.debug("Fetched manifest for %s hash %r", tf, myHash)
                for line in mfh:
                    partList.append(line[:-1])
            #
            logger.debug("Parts (%d) %r", len(partList), partList)
            for part in partList:
                localPath = os.path.join(localDirPath, part)
                remotePath = os.path.join(repoDirPath, bundleFileName[:-7], part)
                logger.debug("%r %r", remotePath, localPath)
                fileU.get(remotePath, localPath)
            #
            sj = SplitJoin()
            ok = sj.join(self.__localStashTarFilePath, localDirPath)
            if ok:
                ok = fileU.unbundleTarfile(self.__localStashTarFilePath, dirPath=localRestoreDirPath)
            return ok
        except Exception as e:
            logger.exception("Failing for %r with %s", bundleFileName, str(e))
            ok = False
        return ok

    def __makeBundleFileName(self, baseBundleFileName, remoteStashPrefix="A"):
        fn = baseBundleFileName
        try:
            fn = "%s-%s" % (remoteStashPrefix.upper(), baseBundleFileName) if remoteStashPrefix else baseBundleFileName
        except Exception as e:
            logger.exception("Failing with %s", str(e))
        return fn



##
# File:    FtpUtil.py
# Author:  dwp
# Date:    3-Aug-2021
#
# Updates:
#
##
"""
Class providing essential data transfer operations for FTP protocol.
"""

__docformat__ = "google en"
__author__ = "Dennis Piehl"
__email__ = "dennis.piehl@rcsb.org"
__license__ = "Apache 2.0"

#
import logging
import os
import ftplib

from rcsb.utils.io.FileUtil import FileUtil

logger = logging.getLogger(__name__)


class FtpUtil(object):
    """Class providing essential data transfer operations for FTP protocol"""

    def __init__(self, *args, **kwargs):
        _ = args
        _ = kwargs
        self.__ftpClient = None
        self.__context = None
        self.__raiseExceptions = False
        #

    # def getRootPath(self):
    #    return self._rootPath

    def connect(self, hostName, userName="anonymous", pw=None, keyFilePath=None, certFilePath=None, certFilePw=None):
        """Establish a connection to an FTP server, using ftplib (https://docs.python.org/3/library/ftplib.html).

        Arguments:
            hostName (str): hostname of FTP server (e.g., 'ftp.rcsb.org')
            userName (str): username to login with
            pw (str): password to login with (if required)
            keyFilePath (str): path to a file containing the private SSL key (for TLS connections)
            certFilePath (str): "path to a single file in PEM format containing the certificate as
                                well as any number of CA certificates needed to establish the
                                certificate’s authenticity" (for TLS connections)
                                (https://docs.python.org/3/library/ssl.html)
            certFilePw (str): password to access the certificate file (if required)

        Returns:
            bool: True for success or false otherwise
        """
        try:
            self.__ftpClient = self.__makeFtpClient(hostName=hostName, userName=userName, pw=pw, keyFilePath=keyFilePath, certFilePath=certFilePath, certFilePw=certFilePw)
            return True
        except Exception as e:
            if self.__raiseExceptions:
                raise e
            else:
                logger.error("Failing connect for hostname %s with %s", hostName, str(e))
                return False

    def __makeFtpClient(self, hostName, userName="anonymous", pw=None, keyFilePath=None, certFilePath=None, certFilePw=None):
        """
        Make an FTP client connected to the specified host using provided authentication parameters
        (username and password for standard FTP connection, and/or certificate and private key files for FTP_TLS connection).

        Arguments:
            hostName (str): hostname of FTP server (e.g., 'ftp.rcsb.org')
            userName (str): username to login with
            pw (str): password to login with (if required)
            keyFilePath (str): path to a file containing the private SSL key (for TLS connections)
            certFilePath (str): "path to a single file in PEM format containing the certificate as
                                well as any number of CA certificates needed to establish the
                                certificate’s authenticity" (for TLS connections)
                                (https://docs.python.org/3/library/ssl.html)
            certFilePw (str): password to access the certificate file (if required)

        Returns:
            ftplib.FTP or ftplib.FTP_TLS:   ftplib FTP connection object, as either a standard FTP connection or with FTP_TLS
                                            (depending if an SSL certificate was provided for the connection or not).
        """
        ftp = None
        self.__context = None
        try:
            if certFilePath is not None:
                # Create SSLContext object
                self.__context = ftplib.ssl.SSLContext()

                # Load the certificate and key (if provided) into the context object
                self.__context.load_cert_chain(certfile=certFilePath, keyfile=keyFilePath, password=certFilePw)

                # Connect via FTP_TLS using supplied method of authentication.
                # For TLS with certificate, host and user may need to be set to empty string '' or None
                ftp = ftplib.FTP_TLS(host=hostName, user=userName, passwd=pw, context=self.__context)
            else:
                # Connect via FTP using username and password authentication
                ftp = ftplib.FTP(host=hostName, user=userName, passwd=pw)

            return ftp
        except Exception as e:
            logger.error("Error occurred creating FTP client: %s: %s", e.__class__, str(e))
            if ftp is not None:
                ftp.close()
            if self.__raiseExceptions:
                raise e

    def mkdir(self, path):
        """Make a directory on a remote FTP server.

        Arguments:
            path (str): directory path

        Returns:
            bool: True for success or false otherwise
        """
        try:
            self.__ftpClient.mkd(path)
            return True
        except Exception as e:
            if self.__raiseExceptions:
                raise e
            else:
                logger.error("mkdir failing for path %s with %s", path, str(e))
                return False

    def dirstat(self, path):
        """Return stat attributes for a given directory path (e.g. ["type", "size", "perm"]).

        Only works if the FTP server supports MLSD (RFC 3659), which not all servers support.
        For a guaranteed way of returning the size of a specific file, use the self.size() method instead.

        Arguments:
            path (str): remote directory path (NOT a file path)

        Returns:
            dict: Dictionary of files and their corresponding stat attributes.
        """
        try:
            dD = {}
            for name, facts in self.__ftpClient.mlsd(path):
                dD.update({name: facts})
            return dD
        except Exception as e:
            if self.__raiseExceptions:
                raise e
            else:
                logger.error("stat failing for path %s with %s", path, str(e))
                return {}

    def size(self, filePath):
        """Return stat attributes for a given file (NOT directory) on a remote FTP server.

        Arguments:
            filePath (str): remote file path

        Returns:
            int: Size of the file in bytes, or None if unsuccessful
        """
        try:
            fSize = self.__ftpClient.size(filePath)
            return int(fSize)
        except Exception as e:
            if self.__raiseExceptions:
                raise e
            else:
                logger.error("stat failing for path %s with %s", filePath, str(e))
                return None

    def put(self, localPath, remotePath):
        """Put a local file on a remote FTP server.

        Arguments:
            localPath (str): local file path
            remotePath (str): remote file path

        Returns:
            bool: True for success or false otherwise
        """
        try:
            # First, make sure the provided localPath represents a file, not a directory
            if not os.path.isfile(localPath):
                logger.error("put failing for localPath %s - path must be to a specific file, not a directory.", localPath)
                return False

            fileU = FileUtil()
            remotePathDir = fileU.getFilePath(remotePath)
            self.mkdir(remotePathDir)
            # If provided remotePath already exists and is a directory, put the file on the remote server using the local filename
            # to avoid unintentionally overwriting an entire remote directory with a single file
            if (os.path.exists(remotePath) and os.path.isdir(remotePath)):
                localFileName = FileUtil().getFileName(localPath)
                remoteFilePath = os.path.join(remotePath, localFileName)
            else:
                remoteFilePath = remotePath
            with open(localPath, 'rb') as lFP:
                self.__ftpClient.storbinary('STOR %s' % remoteFilePath, lFP)
            if remoteFilePath in self.listdir(remotePathDir):
                return True
            else:
                logger.error("put failing for localPath %s remoteFilePath %s", localPath, remoteFilePath)
                return False
        except Exception as e:
            if self.__raiseExceptions:
                raise e
            else:
                logger.error("put failing for localPath %s  remotePath %s with %s", localPath, remotePath, str(e))
                return False

    def get(self, remotePath, localPath):
        """Get a file from a remote FTP server.

        Arguments:
            remotePath (str): remote file path
            localPath (str): local file path

        Returns:
            bool: True for success or false otherwise
        """
        try:
            fileU = FileUtil()
            fileU.mkdirForFile(localPath)
            # If provided localPath already exists and is a directory, retrieve the file using the name on the remote server
            # to avoid unintentionally overwriting an entire local directory with a single retrieved file
            if (os.path.exists(localPath) and os.path.isdir(localPath)):
                remoteFileName = FileUtil().getFileName(remotePath)
                localFilePath = os.path.join(localPath, remoteFileName)
            else:
                localFilePath = localPath
            with open(localFilePath, 'wb') as lFP:
                self.__ftpClient.retrbinary('RETR %s' % remotePath, lFP.write)
            ok = fileU.exists(localFilePath)
            if ok:
                return True
            else:
                logger.error("get failing for remotePath %s localFilePath %s", remotePath, localFilePath)
                return False
        except Exception as e:
            if self.__raiseExceptions:
                raise e
            else:
                logger.error("get failing for remotePath %s localPath %s with %s", remotePath, localPath, str(e))
                return False

    def listdir(self, path):
        """Return a flat list of all files and directories (as full path names) for a given directory.

        E.g., for path='/pub/pdb/data', returns ['/pub/pdb/data/.', '/pub/pdb/data/..', '/pub/pdb/data/biounit', '/pub/pdb/data/bird',
                                                '/pub/pdb/data/component-models', '/pub/pdb/data/monomers', '/pub/pdb/data/status', '/pub/pdb/data/structures']

        Note that not all FTP servers allow the checking of specific file existence, so must provide directory path as input.

        Arguments:
            path (str): directory path

        Returns:
            list: List of all files and directories (as full path names) for a given directory.

        """
        try:
            return self.__ftpClient.nlst(path)
        except Exception as e:
            if self.__raiseExceptions:
                raise e
            else:
                logger.error("listdir failing for path %s with %s", path, str(e))
                return False

    def rmdir(self, dirPath):
        """Remove a directory from a remote FTP server.

        Arguments:
            dirPath (str): remote directory path

        Returns:
            bool: True for success or false otherwise
        """
        try:
            self.__ftpClient.rmd(dirPath)
            return True
        except Exception as e:
            if self.__raiseExceptions:
                raise e
            else:
                logger.error("rmdir failing for path %s with %s", dirPath, str(e))
                return False

    def remove(self, filePath):
        """Remove a file from a remote FTP server.

        Arguments:
            filePath (str): remote file path

        Returns:
            bool: True for success or false otherwise
        """
        try:
            self.__ftpClient.delete(filePath)
            return True
        except Exception as e:
            if self.__raiseExceptions:
                raise e
            else:
                logger.error("remove failing for path %s with %s", filePath, str(e))
                return False

    def close(self):
        """Close a remote FTP connection.

        Returns:
            bool: True for success or false otherwise
        """
        try:
            if self.__ftpClient is not None:
                self.__ftpClient.close()
            return True
        except Exception as e:
            if self.__raiseExceptions:
                raise e
            else:
                logger.error("Close failing with %s", str(e))



##
# File: LogUtil.py
# Date: 5-Sep-2021
#
# Updates:
#
##
import copy
import json
import logging
from datetime import datetime

from rcsb.utils.io.IoUtil import JsonTypeEncoder


class StructFormatter(logging.Formatter):
    """Structured log formatter including message, time, and any extra attributes.

    Example::

        import logging

        from rcsb.utils.io.LogUtil import StructFormatter
        sl = logging.StreamHandler()
        sl.setFormatter(StructFormatter())
        logger = logging.getLogger(__name__)
        logger.addHandler(sl)

    """

    def __init__(self, fmt, mask):
        super(StructFormatter, self).__init__(fmt, mask)
        self._keywords = [
            "args",
            "asctime",
            "created",
            "exc_info",
            "exc_text",
            "filename",
            "funcName",
            "levelname",
            "levelno",
            "lineno",
            "message",
            "module",
            "msecs",
            "msg",
            "name",
            "pathname",
            "process",
            "processName",
            "relativeCreated",
            "stack_info",
            "thread",
            "threadName",
        ]

    def format(self, record):
        """Serialize the log record in JSON including message, time, and extra attributes.
        Exception and stack traceback details are also included when set.

        Args:
            record (obj): log record object (LogRecord)

        Returns:
            (JSON): JSON serialized log record

        """

        myRecord = copy.copy(record)
        message = myRecord.getMessage()

        rD = {ky: myRecord.__dict__[ky] for ky in myRecord.__dict__ if ky not in self._keywords}
        rD["message"] = message
        if "time" not in rD:
            rD["time"] = datetime.utcnow()
        if myRecord.exc_info:
            rD["exc_info"] = self.formatException(myRecord.exc_info)
        if myRecord.stack_info:
            rD["stack_info"] = self.formatStack(myRecord.stack_info)
        #
        return self._serialize(rD)

    def _serialize(self, rD):
        """ """
        try:
            return json.dumps(rD, cls=JsonTypeEncoder)
        except TypeError:
            try:
                return json.dumps(rD)
            except TypeError:
                return "{}"


class DetailedStructFormatter(StructFormatter):
    """Structured logging formatter atomizing all log record attributes and extra attributes.
    Exception and stack traceback details are also included when set.

    Example:

        import logging
        from rcsb.utils.io.LogUtil import DetailedStructFormatter
        sl = logging.StreamHandler()
        sl.setFormatter(DetailedStructFormatter())
        logger = logging.getLogger(__name__)
        logger.addHandler(sl)

    """

    def format(self, record):
        """Serialize the log record in JSON including all standard log record attributes and extra attributes.

        Args:
            record (obj): log record object (LogRecord)

        Returns:
            (JSON): JSON serialized log record
        """
        myRecord = copy.copy(record)
        message = myRecord.getMessage()
        rD = {ky: myRecord.__dict__[ky] for ky in myRecord.__dict__ if ky not in self._keywords}
        rD["message"] = message
        if "time" not in rD:
            rD["time"] = datetime.utcnow()
        if myRecord.exc_info:
            rD["exc_info"] = self.formatException(myRecord.exc_info)
        if myRecord.stack_info:
            rD["stack_info"] = self.formatStack(myRecord.stack_info)
        #
        rD["filename"] = myRecord.filename
        rD["funcName"] = myRecord.funcName
        rD["levelname"] = myRecord.levelname
        rD["lineno"] = myRecord.lineno
        rD["module"] = myRecord.module
        rD["name"] = myRecord.name
        rD["pathname"] = myRecord.pathname
        rD["process"] = myRecord.process
        rD["processName"] = myRecord.processName
        rD["thread"] = myRecord.thread
        rD["threadName"] = myRecord.threadName
        #
        return self._serialize(rD)



##
# File:    DictMethodResourceProvider.py
# Author:  J. Westbrook
# Date:    3-Jun-2019
# Version: 0.001 Initial version
#
#
# Updates:
#
##
"""
Skeleton Singleton metaclass.

"""
__docformat__ = "google en"
__author__ = "John Westbrook"
__email__ = "jwest@rcsb.rutgers.edu"
__license__ = "Apache 2.0"


class _SingletonClass(type):
    __classInstances = {}

    def __call__(cls, *args, **kwargs):
        if cls not in cls.__classInstances:
            cls.__classInstances[cls] = super(_SingletonClass, cls).__call__(*args, **kwargs)
        return cls.__classInstances[cls]

    def clear(cls):
        try:
            del _SingletonClass.__classInstances[cls]
        except KeyError:
            pass


class SingletonClass(_SingletonClass("MySingletonClass", (object,), {})):
    pass



##
# File: CacheUtils.py
# Date: 18-Jul-2019 jdw
#
##

import logging
from collections import OrderedDict

logger = logging.getLogger(__name__)


class CacheUtils:
    """LRU style cache.
    """

    def __init__(self, size=10, label="cache"):
        """LRU style cache

        Args:
            size (int, optional): maximum number of elements in the cache. Defaults to 10.
            label (str, optional): A label an instance of the cache.  Defaults to "cache".
        """
        self.__size = size
        self.__cache = OrderedDict()
        self.__label = label

    def get(self, key):
        """Return the cached value associated with the input key.  The key:value
        combination are moved to top of the cache.

        Args:
            key (hashable): identifier for cached object

        Returns:
            (any): Cached object associated with the input key or None
        """
        try:
            value = self.__cache.pop(key)
            self.__cache[key] = value
            return value
        except KeyError:
            return None

    def set(self, key, value):
        """Store the input value in the cache. The cache contents of the cache
        are adjusted if the size limit is exceeded.

        Args:
            key (hashable): identifier for cached object
            value (any): value to be cached

        Returns:
            bool: True for success or false otherwise
        """
        try:
            try:
                self.__cache.pop(key)
            except KeyError:
                if len(self.__cache) >= self.__size:
                    self.__cache.popitem(last=False)
            self.__cache[key] = value
            return True
        except Exception as e:
            logger.exception("Failing for %s with %s", self.__label, str(e))
        return False



##
# File: FileUtil.py
#
# Skeleton implementation of common File I/O operations.
#
# Updates:
#  5-Jun-2018 jdw add support for local copy operations using shutil.copy
#  6-Mar-2019 jdw add previously stubbed remote access and tar file member support
# 10-Mar-2019 jdw add exists() method
# 11-Aug-2019 jdw incorporate compress/uncompress and toAscii methods from IoUtil.py
#                 add __extractZipMember() method.  Add retry support for __fetchUrl().
#                 Replace urlibx version of __fetchUrl with a version that depends on
#                 requests module to better support redirection and authentication.
##

__docformat__ = "google en"
__author__ = "John Westbrook"
__email__ = "jwest@rcsb.rutgers.edu"
__license__ = "Apache 2.0"

#
import bz2
import contextlib
import gzip
import hashlib
import io
import logging
import os
import shutil
import tarfile
import tempfile
import traceback
import zipfile
import lzma

import requests
from rcsb.utils.io.decorators import retry

# pylint: disable=ungrouped-imports
try:
    import urllib.request as myurl
except Exception:
    import urllib2 as myurl

try:
    from urllib.parse import urlsplit
except Exception:
    from urlparse import urlsplit


logger = logging.getLogger(__name__)


class FileUtil(object):
    """Skeleton implementation for File I/O operations"""

    def __init__(self, workPath=None, **kwargs):
        _ = kwargs
        self.__workPath = workPath
        self.__timeout = None
        if self.__workPath and self.__workPath != ".":
            self.mkdir(workPath)

    def exists(self, locator, mode=os.R_OK):
        """Return if the input file exists with optional mode settings

        Args:
            locator (string): local file path or url
            mode (mode, optional): local mode setting. Defaults to os.R_OK.

        Returns:
            bool: True if file exists or False otherwise
        """
        try:
            if not locator:
                logger.warning("Null locator: %s", "".join(traceback.format_stack()))
                return False
            localFlag = self.isLocal(locator)
            if localFlag:
                return os.access(locator, mode)
            else:
                if locator.startswith("ftp:"):
                    logger.warning("ftp:// protocol not supported.")
                else:
                    response = requests.head(locator, timeout=self.__timeout)
                    logger.debug("response code %r", response.status_code)
                    if response.status_code == 200 and "content-length" in response.headers and int(response.headers["content-length"]) > 0:
                        return True
        except Exception as e:
            logger.exception("Failing for %r with %s", locator, str(e))
        return False

    def size(self, locator):
        """Return the file size in bytes

        Args:
            locator (str): local file path or url

        Returns:
            int: approximate file size in bytes or zero
        """
        try:
            if not locator:
                logger.warning("Null locator: %s", "".join(traceback.format_stack()))
                return False
            localFlag = self.isLocal(locator)
            if localFlag:
                st = os.stat(locator)
                return st.st_size
            else:
                if locator.startswith("ftp:"):
                    logger.warning("ftp:// protocol not supported.")
                else:
                    response = requests.head(locator, timeout=self.__timeout)
                    if response.status_code == 200 and "content-length" in response.headers:
                        return int(response.headers["content-length"])
        except Exception:
            return 0

    def hash(self, filePath, hashType="md5"):
        try:
            localFlag = self.isLocal(filePath)
            if not localFlag:
                return None
            with open(filePath, "rb") as ifh:
                if hashType == "md5":
                    fileHash = hashlib.md5()
                elif hashType == "sha256":
                    fileHash = hashlib.sha256()
                else:
                    logger.error("Unsupported hash type %r", hashType)
                    return None
                chunk = ifh.read(8192)
                while chunk:
                    fileHash.update(chunk)
                    chunk = ifh.read(8192)
            return fileHash.hexdigest()
        except Exception:
            return None

    def mkdir(self, dirPath, mode=0o755):
        """Create the input directory.

        Args:
            dirPath (string): local directory path
            mode (mode, optional): mode setting. Defaults to os.W_OK.

        Returns:
            bool: True for success or False otherwise
        """
        try:
            logger.debug("Checking target directory %s", dirPath)
            if not os.access(dirPath, os.W_OK):
                logger.debug("Creating cache directory %s", dirPath)
                os.makedirs(dirPath, mode)
            return True
        except Exception as e:
            logger.exception("Failing for %s with %s", dirPath, str(e))
        return False

    def mkdirForFile(self, filePath, mode=0o755):
        """Create directory path for the input path if it does not exists.

        Args:
            filePath (string): local file path target
            mode (mode, optional): mode setting. Defaults to os.W_OK.

        Returns:
            bool: True for success or False otherwise
        """
        try:
            logger.debug("Checking target directory for file %s", filePath)
            dirPath, _ = os.path.split(filePath)
            if dirPath:
                return self.mkdir(dirPath, mode)
            else:
                return True
        except Exception as e:
            logger.exception("Failing for %s with %s", filePath, str(e))
        return False

    #
    def remove(self, pth):
        """Method to remove input file, link  or directory."""
        ok = False
        try:
            localFlag = self.isLocal(pth)
            if not localFlag:
                return ok
            if pth and not self.exists(pth):
                ok = True
            if os.path.isfile(pth) or os.path.islink(pth):
                os.unlink(pth)
                ok = True
            elif os.path.isdir(pth):
                shutil.rmtree(pth, True)
                ok = True
        except Exception as e:
            logger.error("Failing for %s with %s", pth, str(e))
            ok = False
        #
        return ok

    def replace(self, srcPath, dstPath):
        """Method to replace input file or directory (local only method)."""
        ok = False
        try:
            localFlag = self.isLocal(srcPath)
            if not localFlag:
                return ok
            if srcPath and dstPath and self.exists(srcPath):
                os.replace(srcPath, dstPath)
                ok = True
        except Exception as e:
            logger.error("Failing for %r -> %r with %s", srcPath, dstPath, str(e))
            ok = False
        #
        return ok

    #
    def get(self, remote, local, **kwargs):
        """Fetch remote file to a local path.

        Arguments:
            remote (TYPE): source locator (remote)
            local (TYPE): destination locator (local)
            **kwargs: Description

                tarMember(str) : target member name in a remote tarfile

        Returns:
            bool: True for success or false otherwise

        """
        try:
            ret = False
            tarMember = kwargs.get("tarMember", None)
            #
            localFlag = self.isLocal(remote)
            #
            if localFlag and not tarMember:
                rPath = self.getFilePath(remote)
                lPath = self.getFilePath(local)
                self.mkdirForFile(lPath)
                shutil.copyfile(rPath, lPath)
                ret = True
            elif localFlag and tarMember:
                # Extract a particular member from a local tar file -
                rPath = self.getFilePath(remote)
                lPath = self.getFilePath(local)
                self.mkdirForFile(lPath)
                ret = self.__extractTarMember(rPath, tarMember, lPath)
                logger.debug("Extract %r from %r to %r status %r", tarMember, rPath, lPath, ret)
            elif not localFlag and tarMember:
                # Extract a particular member from a remote tar file
                tarPath = os.path.join(self.__workPath, self.getFileName(remote))
                ret = self.__fetchUrl(remote, tarPath, **kwargs)
                logger.debug("Fetched %r to %r status %r", remote, tarPath, ret)
                #
                ret = self.__extractTarMember(tarPath, tarMember, self.getFilePath(local)) if ret else False
                logger.debug("Extract %r from %r to %r status %r", tarMember, tarPath, self.getFilePath(local), ret)
            elif not localFlag and not tarMember:
                ret = self.__fetchUrl(remote, self.getFilePath(local), **kwargs)
                logger.debug("Fetched %r localpath %r status %r", remote, self.getFilePath(local), ret)
            else:
                ret = False
            #
        except Exception as e:
            logger.exception("For remote %r local %r failing with %s", remote, local, str(e))

        #
        return ret

    def put(self, local, remote, **kwargs):
        """Copy local file to remote location."""
        _ = kwargs
        try:
            ret = False
            localFlag = self.isLocal(remote)
            #
            if localFlag:
                rPath = self.getFilePath(remote)
                self.mkdirForFile(rPath)
                lPath = self.getFilePath(local)
                shutil.copyfile(lPath, rPath)
                ret = True
            else:
                ret = False
        except Exception as e:
            logger.exception("For remote %r local %r failing with %s", remote, local, str(e))
            ret = False
        return ret

    def getFilePath(self, locator):
        try:
            locSp = urlsplit(locator)
            return locSp.path
        except Exception as e:
            logger.exception("For locator %r failing with %s", locator, str(e))
        return None

    def getFileName(self, locator):
        try:
            locSp = urlsplit(locator)
            (_, fn) = os.path.split(locSp.path)
            return fn
        except Exception as e:
            logger.exception("For locator %r failing with %s", locator, str(e))
        return None

    def isLocal(self, locator):
        try:
            locSp = urlsplit(locator)
            return locSp.scheme in ["", "file"]
        except Exception as e:
            logger.exception("For locator %r failing with %s", locator, str(e))
        return None

    def getScheme(self, locator):
        try:
            locSp = urlsplit(locator)
            return locSp.scheme
        except Exception as e:
            logger.exception("For locator %r Failing with %s", locator, str(e))
        return None

    def bundleTarfile(self, tarFilePath, dirPathList, mode="w:gz", recursive=True):
        """Create a tar file bundle of the contents of the input directory path.

        Args:
            tarFilePath (str): output tar file path
            dirPathList (list): target directory path list to store
            mode (str, optional): the file mode for the tar file. Defaults to "w:gz".
            recursive (bool, optional): include subdirectories recursively. Defaults to True.

        Returns:
            bool: True for success or False otherwise
        """
        ret = True
        curDir = os.getcwd()
        try:
            self.mkdirForFile(tarFilePath)
            with tarfile.open(tarFilePath, mode=mode) as archive:
                for dirPath in dirPathList:
                    dp, tp = os.path.split(dirPath)
                    os.chdir(dp)
                    archive.add(tp, recursive=recursive)
        except Exception as e:
            logger.exception("Failing with %s", str(e))
            ret = False
        os.chdir(curDir)
        return ret

    def unbundleTarfile(self, tarFilePath, dirPath="."):
        """Unbundle input tar file.

        Args:
            tarFilePath (str): path to the input tar bundle file
            dirPath (str, optional): directory path to write extracted data. Defaults to ".".

        Returns:
            bool: True for success or False otherwise
        """
        #   import tarfile contents into dirPath -
        ret = True
        try:

            with tarfile.open(tarFilePath) as tar:
                tar.extractall(path=dirPath)
        except Exception as e:
            logger.exception("Failing with %s", str(e))
            ret = False
        return ret

    def unbundleZipfile(self, zipFilePath, dirPath="."):
        """Unbundle input ZIP file.

        Args:
            zipFilePath (str): path to the input tar bundle file
            dirPath (str, optional): directory path to write extracted data. Defaults to ".".

        Returns:
            bool: True for success or False otherwise
        """
        #   import zipfile contents into dirPath -
        ret = True
        try:
            with zipfile.ZipFile(zipFilePath, mode="r") as zObj:
                zObj.extractall(path=dirPath)
        except Exception as e:
            logger.exception("Failing with %s", str(e))
            ret = False
        return ret

    def __extractTarMember(self, tarFilePath, memberName, memberPath):
        ret = True
        try:
            with tarfile.open(tarFilePath) as tar:
                fIn = tar.extractfile(memberName)
                with open(memberPath, "wb") as ofh:
                    ofh.write(fIn.read())
        except Exception as e:
            logger.exception("Failing with %s", str(e))
            ret = False
        return ret

    def __extractZipMember(self, zipFilePath, memberName, memberPath):
        ret = False
        try:
            with zipfile.ZipFile(zipFilePath, mode="r") as zObj:
                memberList = zObj.namelist()
                for member in memberList:
                    if member == memberName:
                        zObj.extract(member, path=memberPath)
                        ret = True
                        break
        except Exception as e:
            logger.exception("Failing with %s", str(e))
            ret = False
        return ret

    def __fetchUrl(self, url, filePath, **kwargs):

        try:
            noRetry = kwargs.get("noRetry", False)
            scheme = self.getScheme(url)
            ok = self.mkdirForFile(filePath)
            if not ok:
                logger.error("Failing to create target directory for file %r", filePath)
                return False
            if scheme in ["ftp"]:
                return self.__fetchUrlPy(url, filePath, **kwargs)
            else:
                if noRetry:
                    return self.__fetchUrlReqNoRt(url, filePath, **kwargs)
                else:
                    return self.__fetchUrlReqRt(url, filePath, **kwargs)
        except Exception as e:
            logger.error("Failing for url %r with %s", url, str(e))
        return False

    @retry((myurl.URLError, myurl.HTTPError), maxAttempts=3, delaySeconds=5, multiplier=3, defaultValue=False, logger=logger)
    def __fetchUrlPy(self, url, filePath, **kwargs):
        """Fetch data from a remote URL and store this in input filePath.

        Args:
            url (str): target URL to fetch
            filePath (str): path to store the data from the remote url
            username (str): basic auth username
            password (str): dasic auth password
            kwargs (dict, optional): other options

        Raises:
            e: any exception

        Returns:
            bool: True for sucess or False otherwise
        """
        try:
            user = kwargs.get("username", None)
            pw = kwargs.get("password", None)
            if user and pw:
                pwMgr = myurl.HTTPPasswordMgrWithDefaultRealm()
                pwMgr.add_password(None, url, user, pw)
                handler = myurl.HTTPBasicAuthHandler(pwMgr)
                opener = myurl.build_opener(handler)
                # Install the opener.
                myurl.install_opener(opener)
            #
            ok = self.mkdirForFile(filePath)
            if not ok:
                logger.error("Failing to create target directory for file %r", filePath)
                return False
            with open(filePath, "wb") as outFile:
                with contextlib.closing(myurl.urlopen(url)) as fp:
                    blockSize = 1024 * 8
                    while True:
                        block = fp.read(blockSize)  # pylint: disable=no-member
                        if not block:
                            break
                        outFile.write(block)
            return True
        except Exception as e:
            logger.error("Failing for %s with %s", filePath, str(e))
            raise e

        return False

    @retry((requests.exceptions.RequestException), maxAttempts=3, delaySeconds=5, multiplier=3, defaultValue=False, logger=logger)
    def __fetchUrlReqRaw(self, url, filePath, **kwargs):
        """Fetch data from a remote URL and store this in input filePath.

        Args:
            url (str): target URL to fetch
            filePath (str): path to store the data from the remote url
            username (str): basic auth username
            password (str): dasic auth password
            kwargs (dict, optional): other options

        Raises:
            e: any exception

        Returns:
            bool: True for sucess or False otherwise

            with open(filename, 'wb') as fd:
                for chunk in r.iter_content(chunk_size=128):
                    fd.write(chunk)


        """
        user = kwargs.get("username", None)
        pw = kwargs.get("password", None)
        #
        try:
            ok = self.mkdirForFile(filePath)
            if not ok:
                logger.error("Failing to create target directory for file %r", filePath)
                return False
            if user and pw:
                with requests.get(url, stream=True, auth=(user, pw), allow_redirects=True, timeout=self.__timeout) as rIn:
                    with open(filePath, "wb") as fOut:
                        shutil.copyfileobj(rIn.raw, fOut)

            else:
                with requests.get(url, stream=True, allow_redirects=True, timeout=self.__timeout) as rIn:
                    with open(filePath, "wb") as fOut:
                        shutil.copyfileobj(rIn.raw, fOut)

            return True
        except Exception as e:
            logger.error("Failing for %s with %s", filePath, str(e))
            raise e

        return False

    @retry((requests.exceptions.RequestException), maxAttempts=3, delaySeconds=5, multiplier=3, defaultValue=False, logger=logger)
    def __fetchUrlReqRt(self, url, filePath, **kwargs):
        return self.__fetchUrlReqNoRt(url, filePath, **kwargs)

    def __fetchUrlReqNoRt(self, url, filePath, **kwargs):
        """Fetch data from a remote URL and store this in input filePath.

        Args:
            url (str): target URL to fetch
            filePath (str): path to store the data from the remote url
            username (str): basic auth username
            password (str): dasic auth password
            kwargs (dict, optional): other options

        Raises:
            e: any exception

        Returns:
            bool: True for sucess or False otherwise

        """
        user = kwargs.get("username", None)
        pw = kwargs.get("password", None)
        chunkSize = kwargs.get("chunkSize", 1024 * 1024 * 3)
        #
        try:
            ok = self.mkdirForFile(filePath)
            if not ok:
                logger.error("Failing to create target directory for file %r", filePath)
                return False
            if user and pw:
                with requests.get(url, stream=True, auth=(user, pw), allow_redirects=True, timeout=self.__timeout) as rIn:
                    if rIn.status_code == requests.codes.ok:  # pylint: disable=no-member
                        with open(filePath, "wb") as fOut:
                            for chunk in rIn.iter_content(chunk_size=chunkSize):
                                fOut.write(chunk)
                    else:
                        logger.error("Fetch %r fails with status %r", url, rIn.status_code)
                        rIn.raise_for_status()
                        return False
            else:
                with requests.get(url, stream=True, allow_redirects=True, timeout=self.__timeout) as rIn:
                    if rIn.status_code == requests.codes.ok:  # pylint: disable=no-member
                        with open(filePath, "wb") as fOut:
                            for chunk in rIn.iter_content(chunk_size=chunkSize):
                                fOut.write(chunk)
                    else:
                        logger.error("Fetch %r fails with status %r", url, rIn.status_code)
                        rIn.raise_for_status()
                        return False

            return True
        except Exception as e:
            logger.error("Failing for %s with %s", filePath, str(e))
            raise e

        return False

    def toAscii(self, inputFilePath, outputFilePath, chunkSize=5000, encodingErrors="ignore"):
        """Encode input file to Ascii and write this to the target output file.   Handle encoding
        errors according to the input settting ('ignore', 'escape', 'xmlcharrefreplace').
        """
        try:
            ok = self.mkdirForFile(outputFilePath)
            if not ok:
                logger.error("Failing to create target directory for file %r", outputFilePath)
                return False
            chunk = []
            with io.open(inputFilePath, "r", encoding="utf-8") as rIn, io.open(outputFilePath, "w", encoding="ascii") as wOut:
                for line in rIn:
                    # chunk.append(line.encode('ascii', 'xmlcharrefreplace').decode('ascii'))
                    chunk.append(line.encode("ascii", encodingErrors).decode("ascii"))
                    if len(chunk) == chunkSize:
                        wOut.writelines(chunk)
                        chunk = []
                wOut.writelines(chunk)
            return True
        except Exception as e:
            logger.error("Failing text ascii encoding for %s with %s", inputFilePath, str(e))
        #
        return False

    def uncompress(self, inputFilePath, outputDir=None):
        """Uncompress the input file if the path name has a recognized compression type file extension.

        Return the path of the uncompressed file (in outDir) or the original input file path.

        """
        try:
            outputDir = outputDir if outputDir else tempfile.gettempdir()
            _, fn = os.path.split(inputFilePath)
            bn, _ = os.path.splitext(fn)
            outputFilePath = os.path.join(outputDir, bn)
            if inputFilePath.endswith(".gz"):
                with gzip.open(inputFilePath, mode="rb") as inpF:
                    with io.open(outputFilePath, "wb") as outF:
                        shutil.copyfileobj(inpF, outF)
            elif inputFilePath.endswith(".bz2"):
                with bz2.open(inputFilePath, mode="rb") as inpF:
                    with io.open(outputFilePath, "wb") as outF:
                        shutil.copyfileobj(inpF, outF)
            elif inputFilePath.endswith(".xz"):
                with lzma.open(inputFilePath, mode="rb") as inpF:
                    with io.open(outputFilePath, "wb") as outF:
                        shutil.copyfileobj(inpF, outF)
            elif inputFilePath.endswith(".zip"):
                with zipfile.ZipFile(inputFilePath, mode="r") as zObj:
                    memberList = zObj.namelist()
                    for member in memberList[:1]:
                        zObj.extract(member, path=outputDir)
                if memberList:
                    outputFilePath = os.path.join(outputDir, memberList[0])
            else:
                outputFilePath = inputFilePath

        except Exception as e:
            logger.exception("Failing uncompress for file %s with %s", inputFilePath, str(e))
        logger.debug("Returning file path %r", outputFilePath)
        return outputFilePath

    def compress(self, inpPath, outPath, compressType="gzip"):
        try:
            if compressType == "gzip":
                with open(inpPath, "rb") as fIn:
                    with gzip.open(outPath, "wb") as fOut:
                        shutil.copyfileobj(fIn, fOut)
                return True
            else:
                logger.error("Unsupported compressType %r", compressType)
        except Exception as e:
            logger.exception("Compressing file %s failing with %s", inpPath, str(e))
        return False



##
# File: StashableBase.py
#
# Base class template that implements a common pattern for to backup and restore
# cache directories to stash storage.
#
# Updates:
# 19-Jul-2021 jdw add git push support
##

__docformat__ = "google en"
__author__ = "John Westbrook"
__email__ = "jwest@rcsb.rutgers.edu"
__license__ = "Apache 2.0"

import logging
import os
import time

from rcsb.utils.io.StashUtil import StashUtil

logger = logging.getLogger(__name__)


class StashableBase(object):
    """Base class implementing a backup and restore methods for cache directories to/from stash storage."""

    def __init__(self, cachePath, dirNameL):
        """Methods implementing backup and restore operations to and from stash storage. Remote stash
        storage is defined using the standard configuration options. Primary and fallback stash servers
        are supported. Git repositories are also supported as stash targets.

        Args:
            cachePath (str): path to directory containing cached directories
            dirNameL (list): list of target directory names in the cache directory for backup and restore operations
        """
        self.__cachePath = cachePath
        self.__dirNameL = dirNameL
        self.__stU = StashUtil(os.path.join(self.__cachePath, "stash"), self.__dirNameL[0])
        #

    def restore(self, cfgOb, configName, remotePrefix=None, useStash=True, useGit=False):
        """Restore the target cache directory from stash storage.

        Args:
            cfgOb (obj): configuration object (ConfigUtil())
            configName (str): configuration section name
            remotePrefix (str, optional): channel prefix. Defaults to None.
            useStash (bool, optional): use "stash" storage services. Defaults to True.
            useGit (bool, optional): use a git repository service. Defaults to False.

        Returns:
            bool: True for success or False otherwise
        """
        ok = False
        if useStash and not ok:
            ok = self.__restoreFromStash(cfgOb, configName, remotePrefix=remotePrefix)
        #
        if useGit and not ok:
            ok = self.__restoreFromGit(cfgOb, configName, remotePrefix=remotePrefix)
        return ok

    def __restoreFromStash(self, cfgOb, configName, remotePrefix=None):
        """Restore the target cache directory from stash storage.

        Args:
            cfgOb (obj): configuration object (ConfigUtil())
            configName (str): configuration section name
            remotePrefix (str, optional): channel prefix. Defaults to None.

        Returns:
            bool: True for success or False otherwise
        """
        ok = False
        try:
            startTime = time.time()
            url = cfgOb.get("STASH_SERVER_URL", sectionName=configName)
            userName = cfgOb.get("_STASH_AUTH_USERNAME", sectionName=configName)
            password = cfgOb.get("_STASH_AUTH_PASSWORD", sectionName=configName)
            basePath = cfgOb.get("_STASH_SERVER_BASE_PATH", sectionName=configName)
            ok = self.__fromStash(url, basePath, userName=userName, password=password, remoteStashPrefix=remotePrefix)
            logger.info("Restored %r data file from stash (%r)", self.__dirNameL, ok)
            if not ok:
                urlFallBack = cfgOb.get("STASH_SERVER_FALLBACK_URL", sectionName=configName)
                ok = self.__fromStash(urlFallBack, basePath, userName=userName, password=password, remoteStashPrefix=remotePrefix)
                logger.info("Restored %r data file from fallback stash (%r)", self.__dirNameL, ok)
            #
            logger.info("Completed stash restore (%r) at %s (%.4f seconds)", ok, time.strftime("%Y %m %d %H:%M:%S", time.localtime()), time.time() - startTime)
        except Exception as e:
            logger.exception("Failing with %s", str(e))
        #
        return ok

    def __restoreFromGit(self, cfgOb, configName, remotePrefix=None):
        """Restore from a partitioned git stash repository via https fetches.

        Args:
            cfgOb (obj): configuration object (ConfigUtil())
            configName (str): configuration section name
            remotePrefix (str, optional): channel prefix. Defaults to None.

        Returns:
            bool: True for success or False otherwise
        """
        ok = False
        try:
            startTime = time.time()
            # accessToken = cfgOb.get("_STASH_GIT_ACCESS_TOKEN", sectionName=configName)
            gitRawHost = cfgOb.get("_STASH_GIT_RAW_SERVER_HOST", sectionName=configName)
            gitRepositoryPath = cfgOb.get("STASH_GIT_REPOSITORY_PATH", sectionName=configName)
            gitBranch = cfgOb.get("STASH_GIT_REPOSITORY_BRANCH", sectionName=configName)
            # maxMegaBytes = cfgOb.get("STASH_GIT_REPOSITORY_MAX_SIZE_MB", sectionName=configName)
            #
            ok = self.__stU.fetchPartitionedBundle(self.__cachePath, gitRepositoryPath, gitRawHost=gitRawHost, gitBranch=gitBranch, remoteStashPrefix=remotePrefix)
            logger.info(
                "Completed git restore for %r (%r) data at %s (%.4f seconds)",
                self.__dirNameL,
                ok,
                time.strftime("%Y %m %d %H:%M:%S", time.localtime()),
                time.time() - startTime,
            )
        except Exception as e:
            logger.exception("Failing with %s", str(e))
        return ok

    def backup(self, cfgOb, configName, remotePrefix=None, useStash=True, useGit=False):
        """Backup the target cache directory to remote stash and/or git storage.

        Args:
            cfgOb (obj): configuration object (ConfigUtil())
            configName (str): configuration section name
            remotePrefix (str, optional): channel prefix. Defaults to None.
            useStash (bool, optional): use "stash" storage services. Defaults to True.
            useGit (bool, optional): use a git repository service. Defaults to False.

        Returns:
            bool: True for success or False otherwise
        """

        ok = self.__stU.makeBundle(self.__cachePath, self.__dirNameL)
        ok1 = True
        if ok and useStash:
            ok1 = self.__backupToStash(cfgOb, configName, remotePrefix=remotePrefix)
        ok2 = True
        if ok and useGit:
            ok2 = self.__backupToGit(cfgOb, configName, remotePrefix=remotePrefix)
        return ok and ok1 and ok2

    def __backupToStash(self, cfgOb, configName, remotePrefix=None):
        """Backup the target cache directory to stash storage.

        Args:
            cfgOb (obj): configuration object (ConfigUtil())
            configName (str): configuration section name
            remotePrefix (str, optional): channel prefix. Defaults to None.


        Returns:
            bool: True for success or False otherwise
        """
        ok1 = ok2 = False
        try:
            startTime = time.time()
            userName = cfgOb.get("_STASH_AUTH_USERNAME", sectionName=configName)
            password = cfgOb.get("_STASH_AUTH_PASSWORD", sectionName=configName)
            basePath = cfgOb.get("_STASH_SERVER_BASE_PATH", sectionName=configName)
            url = cfgOb.get("STASH_SERVER_URL", sectionName=configName)
            urlFallBack = cfgOb.get("STASH_SERVER_FALLBACK_URL", sectionName=configName)
            ok1 = self.__toStash(url, basePath, userName=userName, password=password, remoteStashPrefix=remotePrefix)
            ok2 = self.__toStash(urlFallBack, basePath, userName=userName, password=password, remoteStashPrefix=remotePrefix)
            logger.info(
                "Completed backup for %r data (%r/%r) at %s (%.4f seconds)",
                self.__dirNameL,
                ok1,
                ok2,
                time.strftime("%Y %m %d %H:%M:%S", time.localtime()),
                time.time() - startTime,
            )
        except Exception as e:
            logger.exception("Failing with %s", str(e))
        return ok1 & ok2

    def __backupToGit(self, cfgOb, configName, remotePrefix=None):
        """Backup the to the git stash repository.

        Args:
            cfgOb (obj): configuration object (ConfigUtil())
            configName (str): configuration section name
            remotePrefix (str, optional): channel prefix. Defaults to None.

        Returns:
            bool: True for success or False otherwise
        """
        ok = False
        try:
            startTime = time.time()
            accessToken = cfgOb.get("_STASH_GIT_ACCESS_TOKEN", sectionName=configName)
            gitHost = cfgOb.get("_STASH_GIT_SERVER_HOST", sectionName=configName)
            gitRepositoryPath = cfgOb.get("STASH_GIT_REPOSITORY_PATH", sectionName=configName)
            gitBranch = cfgOb.get("STASH_GIT_REPOSITORY_BRANCH", sectionName=configName)
            maxMegaBytes = cfgOb.get("STASH_GIT_REPOSITORY_MAX_SIZE_MB", sectionName=configName)
            #
            ok = self.__stU.pushBundle(gitRepositoryPath, accessToken, gitHost=gitHost, gitBranch=gitBranch, remoteStashPrefix=remotePrefix, maxSizeMB=maxMegaBytes)
            logger.info(
                "Completed git backup for %r (%r) data at %s (%.4f seconds)",
                self.__dirNameL,
                ok,
                time.strftime("%Y %m %d %H:%M:%S", time.localtime()),
                time.time() - startTime,
            )
        except Exception as e:
            logger.exception("Failing with %s", str(e))
        return ok

    def __toStash(self, url, stashRemoteDirPath, userName=None, password=None, remoteStashPrefix=None):
        """Copy tar and gzipped bundled cache data to remote server/location.

        Args:
            url (str): server URL (e.g. sftp://hostname.domain) None for local host
            stashRemoteDirPath (str): path to target directory on remote server
            userName (str, optional): server username. Defaults to None.
            password (str, optional): server password. Defaults to None.
            remoteStashPrefix (str, optional): channel prefix. Defaults to None.

        Returns:
            (bool): True for success or False otherwise
        """
        ok = False
        try:
            if not url:
                # stashing locally then do this under the cache path
                stashRemoteDirPath = os.path.join(self.__cachePath, stashRemoteDirPath)
            ok = self.__stU.storeBundle(url, stashRemoteDirPath, remoteStashPrefix=remoteStashPrefix, userName=userName, password=password)
        except Exception as e:
            logger.error("Failing with url %r stashDirPath %r: %s", url, stashRemoteDirPath, str(e))
        return ok

    def __fromStash(self, url, stashRemoteDirPath, userName=None, password=None, remoteStashPrefix=None):
        """Restore local cache from a tar and gzipped bundle to fetched from a remote server/location.

        Args:
            url (str): server URL (e.g. sftp://hostname.domain) None for local host
            stashRemoteDirPath (str): path to target directory on remote server
            userName (str, optional): server username. Defaults to None.
            password (str, optional): server password. Defaults to None.
            remoteStashPrefix (str, optional): channel prefix. Defaults to None.

        Returns:
            (bool): True for success or False otherwise
        """
        ok = False
        try:
            if not url:
                stashRemoteDirPath = os.path.join(self.__cachePath, stashRemoteDirPath)
            ok = self.__stU.fetchBundle(self.__cachePath, url, stashRemoteDirPath, remoteStashPrefix=remoteStashPrefix, userName=userName, password=password)
        except Exception as e:
            logger.error("Failing with url %r stashDirPath %r: %s", url, stashRemoteDirPath, str(e))
        return ok



##
# File: GitUtil.py
# Date: 17-Jul-2021
#
# Simple git utilities to perform status, clone, add, push and pull operations.
#
# Updates:
#
##

__docformat__ = "google en"
__author__ = "John Westbrook"
__email__ = "jwest@rcsb.rutgers.edu"
__license__ = "Apache 2.0"

# pylint: disable=assigning-non-slot
#
import logging

from git import Repo

logger = logging.getLogger(__name__)


class GitUtil(object):
    """Simple git utilities to perform status, clone, add, push and pull operations."""

    def __init__(self, token=None, repositoryHost="github.com"):
        """Set GitHub personal access token and repository host.

        Args:
            token (str,optional): GitHub personal access token. Defaults to None.
            repositoryHost (str, optional): repository host name.  Defaults to github.com.
        """
        self.__token = token
        self.__repositoryHost = repositoryHost

    def clone(self, repositoryPath, localRepositoryPath, branch="master"):
        """Clone a GitHub repository to a local repository path.

        Args:
            repositoryPath (str): GitHub repository path (rcsb/py-mmcif)
            localRepositoryPath (str): local repository directory path

        Returns:
            (bool):  True for success or False otherwise
        """
        try:
            remote = f"https://{self.__token}@{self.__repositoryHost}/{repositoryPath}" if self.__token else f"https://{self.__repositoryHost}/{repositoryPath}"
            Repo.clone_from(remote, localRepositoryPath, branch=branch)
        except Exception as e:
            logger.exception("Failing for %r %r (%r) with %s", repositoryPath, localRepositoryPath, branch, str(e))
        return True

    def status(self, localRepositoryPath, branch="master"):
        """Get repository status.

        Args:
            localRepositoryPath (str): local repository directory path

        Returns:
            str: status message
        """
        repo = Repo(localRepositoryPath)
        return repo.git.status(branch)

    def addAll(self, localRepositoryPath, branch="master", remote="origin"):
        """Add/stage any unstaged artifact(s) to the repository (git add -a).

        Args:
            localRepositoryPath (str): local repository directory path
            branch (str, optional): branch name. Defaults to "master".
            remote (str, optional): remote name. Defaults to "origin".

        Returns:
            bool: True for success or False otherwise
        """
        try:
            repo = Repo(localRepositoryPath)
            remoteBranches = list(repo.branches)
            if branch not in remoteBranches:
                origin = repo.remote(name=remote)
                repo.head.reference = repo.create_head(branch)
                repo.head.reference.set_tracking_branch(origin.refs.master)
            repo.git.checkout(branch)
            logger.info("Active branch: %s", repo.active_branch)
            # repo.git.add(update=True)
            repo.git.add(all=True)
            stagedFiles = repo.index.diff("HEAD")
            logger.info("Staged files: %r", [sf.a_path for sf in stagedFiles])
            ok = len(stagedFiles) > 0
            if repo.untracked_files:
                logger.info("Untracked files: %r", repo.untracked_files)
                ok = False
            return ok
        except Exception as e:
            logger.exception("Failing for %r (%r) with %s", localRepositoryPath, branch, str(e))
        return False

    def commit(self, localRepositoryPath, commitMessage="GitUtil automated update", branch="master"):
        """Commit staged changes to the repository.

        Args:
            localRepositoryPath (str): local repository directory path
            commitMessage (str, optional): commit message. Defaults to "GitUtil automated update".
            branch (str, optional): branch name. Defaults to "master".

        Returns:
            bool: True for success or False otherwise
        """
        try:
            repo = Repo(localRepositoryPath)
            repo.git.checkout(branch)
            logger.info("Branch %s", repo.active_branch)
            commitRet = repo.index.commit(commitMessage)
            logger.debug("Commit status %r", commitRet)
            return True
        except Exception as e:
            logger.exception("Failing commit for %r (%r) with %s", localRepositoryPath, branch, str(e))
        return False

    def push(self, localRepositoryPath, remote="origin", branch="master"):
        """Push staged changes to the remote repository service.

        Args:
            localRepositoryPath (str): local repository directory path
            remote (str, optional): remote name. Defaults to "origin".
            branch (str, optional): branch name. Defaults to "master".

        Returns:
             bool: True for success or False otherwise
        """
        #
        try:
            repo = Repo(localRepositoryPath)
            origin = repo.remote(name=remote)
            pushRet = origin.push(branch)[0]
            logger.debug("Push returns %s", pushRet.summary.strip())
            return True
        except Exception as e:
            logger.exception("Failing push for %r (%r) with %s", localRepositoryPath, branch, str(e))
        return False

    def pull(self, localRepositoryPath, remote="origin", branch="master"):
        """Pull changes from the remote repository to the local working copy.

        Args:
            localRepositoryPath (str): local repository directory path
            remote (str, optional): remote name. Defaults to "origin".
            branch (str, optional): branch name. Defaults to "master".

        Returns:
             bool: True for success or False otherwise
        """
        #
        try:
            repo = Repo(localRepositoryPath)
            origin = repo.remote(name=remote)
            fInfo = origin.pull(branch)[0]
            fS = fInfo.note.strip()
            if fS:
                logger.info("Pull returns %s", fInfo.note.strip())
            return True
        except Exception as e:
            logger.exception("Failing push for %r (%r) with %s", localRepositoryPath, branch, str(e))
        return False



##
# File: decorators.py
# Date:  9-Aug-2019 Jdw
#

import multiprocessing
import signal
import sys
import time
from functools import wraps


def quietHook(kind, message, traceback):
    if QuietException in kind.__bases__:
        print("{0}: {1}".format(kind.__name__, message))  # Only print Error Type and Message
    else:
        sys.__excepthook__(kind, message, traceback)  # Print Error Type, Message and Traceback


sys.excepthook = quietHook


class QuietException(Exception):
    pass


class TimeoutException(Exception):
    def __init__(self, value):
        super(TimeoutException, self).__init__(value)
        self.value = value

    def __str__(self):
        return str(self.value)

    def __repr__(self):
        return repr(self.value)


class MyProcess(multiprocessing.Process):
    def __init__(self, func, *args, **kwargs):
        self.queue = multiprocessing.Queue(maxsize=1)
        args = (func,) + args
        multiprocessing.Process.__init__(self, target=self.runner, args=args, kwargs=kwargs)

    def runner(self, func, *args, **kwargs):
        try:
            result = func(*args, **kwargs)
            self.queue.put((True, result))
        except Exception as e:
            self.queue.put((False, e))

    def done(self):
        return self.queue.full()

    def result(self):
        return self.queue.get()


def timeout(seconds, message="Function call timed out"):
    def wrapper(function):
        def _handleTimeout(signum, frame):
            raise TimeoutException(message)

        @wraps(function)
        def wrapped(*args, **kwargs):
            signal.signal(signal.SIGALRM, _handleTimeout)
            signal.alarm(seconds)
            try:
                result = function(*args, **kwargs)
            finally:
                signal.alarm(0)
            return result

        return wrapped

    return wrapper


def timeoutMp(seconds, forceKill=True):
    def wrapper(function):
        @wraps(function)
        def wrapped(*args, **kwargs):
            now = time.time()
            proc = MyProcess(function, *args, **kwargs)
            proc.start()
            proc.join(seconds)
            if proc.is_alive():
                if forceKill:
                    proc.terminate()
                runtime = int(time.time() - now)
                raise TimeoutException("timed out after {0} seconds".format(runtime))
            assert proc.done()
            success, result = proc.result()
            if success:
                return result
            else:
                raise result

        return wrapped

    return wrapper


def retry(targetException, maxAttempts=3, delaySeconds=2, multiplier=3, defaultValue=None, logger=None):
    """Retry the method or function on exception after a delay interval which grows by a
    multiplicative factor between attempts.  On failure after maxAttempts are exceeded then default value
    is returned.

    Args:
        targetException (Exception or tuple of exceptions): target exception
        maxAttempts (int, optional): number of attempts. Defaults to 3.
        delaySeconds (int, optional): interval to wait between attempts. Defaults to 2.
        multiplier (int, optional): multiplicative factor to apply to delay interval. Defaults to 3.

    Adapted from example at: http://code.activestate.com/recipes/580745-retry-decorator-in-python/
                         and https://wiki.python.org/moin/PythonDecoratorLibrary#Retry
    """

    def wrapper(function):
        @wraps(function)
        def wrapped(*args, **kwargs):
            attemptsLeft = maxAttempts
            currentDelaySecs = delaySeconds
            while attemptsLeft > 1:
                try:
                    if logger:
                        logger.debug("Attempts left %d", attemptsLeft)
                    return function(*args, **kwargs)
                except targetException as e:
                    if logger:
                        logger.info("waiting for %d seconds after %s", currentDelaySecs, str(e))
                    time.sleep(currentDelaySecs)
                    attemptsLeft -= 1
                    currentDelaySecs *= multiplier
            try:
                return function(*args, **kwargs)
            except targetException:
                if logger:
                    logger.info("Failing after %d attempts returning default %r", maxAttempts, defaultValue)
                return defaultValue
            return

        return wrapped

    return wrapper



import os

from rcsb.utils.io.FileUtil import FileUtil


class SplitJoin(object):
    def __init__(self, *kwargs):
        pass

    def split(self, inputFilePath, splitDirPath, prefixName="part_", maxSizeMB=50):
        chunkSize = maxSizeMB * 1000000
        partNumber = 0
        fU = FileUtil()
        fU.mkdir(splitDirPath)
        manifestPath = os.path.join(splitDirPath, "MANIFEST")
        myHash = fU.hash(inputFilePath, hashType="md5")
        with open(manifestPath, "w") as mfh:
            mfh.write("%s\t%s\n" % (inputFilePath, myHash))
            with open(inputFilePath, "rb") as ifh:
                chunk = ifh.read(chunkSize)
                while chunk:
                    partNumber += 1
                    partName = prefixName + str(partNumber)
                    fp = os.path.join(splitDirPath, partName)
                    with open(fp, "wb") as ofh:
                        ofh.write(chunk)
                    mfh.write("%s\n" % partName)
                    #
                    chunk = ifh.read(chunkSize)
        return partNumber

    def join(self, outputFilePath, splitDirPath):
        manifestPath = os.path.join(splitDirPath, "MANIFEST")
        with open(outputFilePath, "wb") as ofh:
            with open(manifestPath, "r") as mfh:
                line = mfh.readline()
                fp, priorHash = line[:-1].split("\t")
                for line in mfh:
                    fp = os.path.join(splitDirPath, line[:-1])
                    with open(fp, "rb") as ifh:
                        data = ifh.read()
                        ofh.write(data)
        fU = FileUtil()
        newHash = fU.hash(outputFilePath, hashType="md5")
        return newHash == priorHash



##
# File:    IndexUtils.py
# Author:  jdw
# Date:    20-June-2016
# Version: 0.001
#
# Updates:
#    22-Jun-2016 jdw add CcdcMatchIndex/Inst() classes -
#    28-Jul-2017 jdw add PdbxChemCompIndex and PdbxChemCompIndexInst classes -
#    28-Jul-2017 jdw change variant_id to target_id in the ccdc results index
#    12-Jan-2021 jdw Py39 cleanup
#    20-Jan-2021 jdw move to rcsb.utils.io and cleanup
##
"""
IndexUtils() implements a simple iterator over a linear index of objects -

"""
__docformat__ = "google en"
__author__ = "John Westbrook"
__email__ = "john.westbrook@rcsb.org"
__license__ = "Apache 2.0"

import logging
import sys
import os

from rcsb.utils.io.MarshalUtil import MarshalUtil

logger = logging.getLogger(__name__)


class IndexBase(object):
    """Base iterator class."""

    def __init__(self, indexFilePath, func, verbose=True):
        self._verbose = verbose
        self._debug = False
        self._rL = []
        self.__func = func
        self.__fmt = "json"
        self._indexFilePath = indexFilePath
        self._indexPath, self._indexFileName = os.path.split(self._indexFilePath)

    def get(self, index=0):
        try:
            return self._rL[index]
        except Exception:
            return []

    def __iter__(self):
        return self.forward()

    def forward(self):
        # Forward generator
        currentRow = 0
        while currentRow < len(self._rL):
            row = self._rL[currentRow]
            currentRow += 1
            yield self.__func(row)

    def reverse(self):
        # The reverse generator
        currentRow = len(self._rL)
        while currentRow > 0:
            currentRow -= 1
            yield self.__func(self._rL[currentRow])

    def clear(self):
        self._rL = []
        return True

    def writeIndex(self):
        try:
            mU = MarshalUtil()
            ok = mU.doExport(self._indexFilePath, self._rL, fmt=self.__fmt, indent=3)
            return ok
        except Exception as e:
            logger.error("Failing with %s", str(e))

        return False

    def readIndex(self):
        try:
            mU = MarshalUtil()
            if not mU.exists(self._indexFilePath):
                return False
            indexObj = mU.doImport(self._indexFilePath, fmt=self.__fmt)
            if indexObj is not None and len(indexObj) > 0:
                self._rL.extend(indexObj)
            return True
        except Exception as e:
            logger.error("Failing with %s", str(e))

        return False

    def load(self, rowList):
        self._rL.extend(rowList)

    def dump(self):
        for ii, dD in enumerate(self._rL):
            logger.info("%4d: %r", ii, dD)
        logger.info("Completed dump")

    def sort(self, atName="identifier", descending=False):
        try:
            self._rL = sorted(self._rL, key=lambda k: k[atName], reverse=descending)
            return True
        except Exception:
            pass
        return False


class CcdcMatchIndex(IndexBase):
    def __init__(self, indexFilePath=None, verbose=True):
        obj = CcdcMatchIndexInst(None, verbose=verbose)
        super(CcdcMatchIndex, self).__init__(indexFilePath, obj.set, verbose)
        #
        self.readIndex()
        if self._debug:
            self.dump()


class CcdcMatchIndexInst(object):
    """Accessor methods for CCDC match index elements."""

    def __init__(self, dObj=None, verbose=True, log=sys.stderr):
        self.__verbose = verbose
        self._lfh = log
        self.__dObj = dObj if dObj is not None else {}

    def __getAttribute(self, name):
        try:
            return self.__dObj[name]
        except Exception:
            return None

    def set(self, dObj=None):
        self.__dObj = dObj
        return self

    def get(self):
        return self.__dObj

    def getCsdVersion(self):
        return self.__getAttribute("csd_version")

    def getCsdDirectory(self):
        return self.__getAttribute("csd_directory")

    def getIdentifier(self):
        return self.__getAttribute("identifier")

    def getChemicalName(self):
        return self.__getAttribute("chemical_name")

    def getTargetId(self):
        return self.__getAttribute("target_id")

    def getTargetPath(self):
        return self.__getAttribute("target_path")

    def getTargetCcPath(self):
        return self.__getAttribute("target_cc_path")

    def getMatchType(self):
        return self.__getAttribute("match_type")

    def getMatchNumber(self):
        return self.__getAttribute("match_number")

    def getMol2Path(self):
        return self.__getAttribute("mol2_file_path")

    def getMolPath(self):
        return self.__getAttribute("mol_file_path")

    #

    def getRFactor(self):
        return self.__getAttribute("r_factor")

    def getTemperature(self):
        return self.__getAttribute("temperature")

    def getRadiationSource(self):
        return self.__getAttribute("radiation_source")

    def getCitationDOI(self):
        return self.__getAttribute("doi")

    def getMatchedAtomLength(self):
        return self.__getAttribute("match_atoms")

    def getHasDisorder(self):
        return self.__getAttribute("has_disorder")

    def getSimilarityScore(self):
        return self.__getAttribute("similarity")

    #
    def setCsdVersion(self, v):
        self.__dObj["csd_version"] = v

    def setCsdDirectory(self, v):
        self.__dObj["csd_directory"] = v

    def setIdentifier(self, v):
        self.__dObj["identifier"] = v

    def setChemicalName(self, name):
        self.__dObj["chemical_name"] = name

    def setTargetId(self, ccId):
        self.__dObj["target_id"] = ccId

    def setTargetPath(self, pth):
        self.__dObj["target_path"] = pth

    def setTargetCcPath(self, pth):
        self.__dObj["target_cc_path"] = pth

    def setMatchType(self, v):
        self.__dObj["match_type"] = v

    def setMatchNumber(self, v):
        self.__dObj["match_number"] = v

    def setMolPath(self, fp):
        self.__dObj["mol_file_path"] = fp

    def setMol2Path(self, fp):
        self.__dObj["mol2_file_path"] = fp

    def setRFactor(self, rv):
        self.__dObj["r_factor"] = rv

    def setTemperature(self, tv):
        self.__dObj["temperature"] = tv

    def setRadiationSource(self, sv):
        self.__dObj["radiation_source"] = sv

    def setSimilarityScore(self, sv):
        self.__dObj["similarity_score"] = sv

    def setMatchedAtomLength(self, lv):
        self.__dObj["match_atoms"] = lv

    def setHasDisorder(self, lv):
        self.__dObj["has_disorder"] = lv

    def setCitationDOI(self, doi):
        self.__dObj["doi"] = doi

    def __repr__(self):
        return str(self.__dObj)

    def __str__(self):
        return str(self.__dObj)



##
# File: MarshalUtil.py
# Date: 4-Jun-2018
#
# Updates:
#      19-Jun-2018 jdw propagate the class workpath to serialize/deserialize methods explicitly
#       6-Mar-2019 jdw add previously stubbed remote access and tar file member support
#       9-Mar-2019 jdw add exists()
#      13-Aug-2019 jdw add multipart support for json/pickle
#
# For py 27 pip install backports.tempfile
##
# pylint: disable=all
__docformat__ = "google en"
__author__ = "John Westbrook"
__email__ = "jwest@rcsb.rutgers.edu"
__license__ = "Apache 2.0"

import logging
import os
import sys

from rcsb.utils.io.FileUtil import FileUtil
from rcsb.utils.io.IoUtil import IoUtil

if sys.version_info[0] == 2:
    from backports import tempfile
else:
    import tempfile


logger = logging.getLogger(__name__)


class MarshalUtil(object):
    """Wrapper for serialization and deserialization methods."""

    def __init__(self, **kwargs):
        self.__workPath = kwargs.get("workPath", ".")
        self.__workDirSuffix = kwargs.get("workDirSuffix", "marshall_")
        self.__workDirPrefix = kwargs.get("workDirSuffix", "_tempdir")
        #
        self.__fileU = FileUtil(workPath=self.__workPath)
        self.__ioU = IoUtil()

    def doExport(self, locator, obj, fmt="list", marshalHelper=None, numParts=None, **kwargs):
        """Serialize the input object at locator path in specified format.  The
        input object is optionally preprocessed by the helper method.

        Args:
            locator (str): target path or URI
            obj (object): data to be serialized
            fmt (str, optional): format for serialization (mmcif, tdd, csv, list). Defaults to "list".
            marshalHelper (method, optional): pre-processor method applied to input data object. Defaults to None.
            numParts (int, optional): serialize the data in parts. Defaults to None. (json and pickle formats)
        Returns:
            bool: True for sucess or False otherwise
        """
        try:
            ret = False
            localFlag = self.__fileU.isLocal(locator)
            if marshalHelper:
                myObj = marshalHelper(obj, **kwargs)
            else:
                myObj = obj
            #
            if localFlag and numParts and fmt in ["json", "pickle"]:
                localFilePath = self.__fileU.getFilePath(locator)
                ret = self.__ioU.serializeInParts(localFilePath, myObj, numParts, fmt=fmt, **kwargs)
            elif localFlag:
                localFilePath = self.__fileU.getFilePath(locator)
                ret = self.__ioU.serialize(localFilePath, myObj, fmt=fmt, workPath=self.__workPath, **kwargs)
            else:
                with tempfile.TemporaryDirectory(suffix=self.__workDirSuffix, prefix=self.__workDirPrefix, dir=self.__workPath) as tmpDirName:
                    # write a local copy then copy to destination -
                    #
                    localFilePath = os.path.join(self.__workPath, tmpDirName, self.__fileU.getFileName(locator))
                    ok1 = self.__ioU.serialize(localFilePath, myObj, fmt=fmt, workPath=self.__workPath, **kwargs)
                    ok2 = True
                    if ok1:
                        ok2 = self.__fileU.put(localFilePath, locator, **kwargs)
                ret = ok1 and ok2
        except Exception as e:
            logger.exception("Exporting locator %r failing with %s", locator, str(e))

        return ret

    def doImport(self, locator, fmt="list", marshalHelper=None, numParts=None, **kwargs):
        """Deserialize data at the target locator in specified format. The deserialized
        data is optionally post-processed by the input helper method.

        Args:
            locator (str): path or URI to input data
            fmt (str, optional): format for deserialization (mmcif, tdd, csv, list). Defaults to "list".
            marshalHelper (method, optional): post-processor method applied to deserialized data object. Defaults to None.
            numParts (int, optional): deserialize the data in parts. Defaults to None. (json and pickle formats)
            tarMember (str, optional): name of a member of tar file bundle. Defaults to None. (tar file format)

        Returns:
            Any: format specific return type
        """
        try:
            tarMember = kwargs.get("tarMember", None)
            localFlag = self.__fileU.isLocal(locator) and not tarMember
            #
            if localFlag and numParts and fmt in ["json", "pickle"]:
                filePath = self.__fileU.getFilePath(locator)
                ret = self.__ioU.deserializeInParts(filePath, numParts, fmt=fmt, **kwargs)
            elif localFlag:
                filePath = self.__fileU.getFilePath(locator)
                ret = self.__ioU.deserialize(filePath, fmt=fmt, workPath=self.__workPath, **kwargs)
            else:
                #
                if fmt == "mmcif":
                    ret = self.__ioU.deserialize(locator, fmt=fmt, workPath=self.__workPath, **kwargs)
                else:
                    with tempfile.TemporaryDirectory(suffix=self.__workDirSuffix, prefix=self.__workDirPrefix, dir=self.__workPath) as tmpDirName:
                        #
                        # Fetch first then read a local copy -
                        #
                        if tarMember:
                            localFilePath = os.path.join(self.__workPath, tmpDirName, tarMember)
                        else:
                            localFilePath = os.path.join(self.__workPath, tmpDirName, self.__fileU.getFileName(locator))

                        # ---  Local copy approach ---
                        self.__fileU.get(locator, localFilePath, **kwargs)
                        ret = self.__ioU.deserialize(localFilePath, fmt=fmt, workPath=self.__workPath, **kwargs)

            if marshalHelper:
                ret = marshalHelper(ret, **kwargs)
        except Exception as e:
            logger.exception("Importing locator %r failing with %s", locator, str(e))
            ret = None
        return ret

    def exists(self, filePath, mode=os.R_OK):
        return self.__fileU.exists(filePath, mode=mode)

    def mkdir(self, dirPath, mode=0o755):
        return self.__fileU.mkdir(dirPath, mode=mode)

    def remove(self, pth):
        return self.__fileU.remove(pth)
